/*-
 * Copyright (C)2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2021 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: del.c,v 1.3 2021/03/14 14:26:08 babolo Exp $"

#define BLIN_COMPAT     3
#define MIFE_COMPAT     5
#define Bpars_COMPAT    4
#define RECOBE_COMPAT   4

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <zlib.h>
#include <babolo/BLINflag.h>
#include <babolo/recobe.h>
#include <babolo/parser.h>
#include <mife.h>
#include "pngtr.h"
#include "headers.h"

void
/*****************************************************************************************************
 *****************************************************************************************************
 ****     png (mode == 0)  img (mode == 1)                              ****
 ****    num                                                               ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_freefile(pngtr_conf *cf, int mode) {                                                       /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_infile     *in = !mode ? &cf->convey[cf->num].png : &cf->convey[cf->num].img;

    ifBLIN_QX3("+ %08X q%u %s", cf->flags, cf->num, !mode ? "png" : "img");
    if  (!in->md) {
        free(in->cont);
    } else {
        mife_fini(in->md);
    }
    free(in->nm);
    bzero(in, sizeof(pngtr_infile));
    ifBLIN_QX3("-");
    return;
#   undef blin_internal_flags
}

void
/*****************************************************************************************************
 *****************************************************************************************************
 ****      opctr  chctr   num                      ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_freeop(pngtr_conf *cf, u_int chctr, u_int opctr) {                                         /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey     *convey;
    pngtr_chunk      *chunk;
    pngtr_operand    *op;

    ifBLIN_QX3("+ %08X q%u c%u o%u", cf->flags, cf->num, chctr, opctr);
    if  (!cf->convey      || !(convey = &cf->convey[cf->num] )) goto out;
    if  (!convey->chunks  || !(chunk = &convey->chunks[chctr])) goto out;
    if  (!chunk->operands || !(op = &chunk->operands[opctr]  )) goto out;
    if  (PNGTR_VIRTUAL & op->flags) free(op->val);
    free(op->select);
out:
    ifBLIN_QX3("-");
    return;
#   undef blin_internal_flags
}

void
/*****************************************************************************************************
 *****************************************************************************************************
 ****      chctr   num                                     ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_freechunk(pngtr_conf *cf, u_int chctr) {                                                   /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey     *convey;
    pngtr_chunk      *chunk;

    ifBLIN_QX3("+ %08X q%u c%u", cf->flags, cf->num, chctr);
    if  (!cf->convey      || !(convey = &cf->convey[cf->num])) goto out;
    if  (!convey->chunks  || !(chunk = &convey->chunks[chctr])) goto out;
    for (u_int opctr = 0; opctr < convey->chunks[chctr].carg; ++opctr) pngtr_freeop(cf, chctr, opctr);
    free(convey->chunks[chctr].operands);
    if  (PNGTR_INFLA & convey->chunks[chctr].flags) free(convey->chunks[chctr].infla);
out:
    ifBLIN_QX3("-");
    return;
#   undef blin_internal_flags
}

void
/*****************************************************************************************************
 *****************************************************************************************************
 ****      num                                                    ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_freeconvey(pngtr_conf *cf) {                                                               /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey     *convey;

    ifBLIN_QX3("+ %08X q%u", cf->flags, cf->num);
    if  (!cf->convey || !(convey = &cf->convey[cf->num])) goto out;
    for (u_int chctr = 0; chctr <= convey->chunum; ++chctr) pngtr_freechunk(cf, chctr);
    pngtr_freefile(cf, 0);
    pngtr_freefile(cf, 1);
    free(convey->imda);
    convey->imda = NULL;
    convey->nowsz = 0;
    free(convey->chunks);
    convey->chunks = NULL;
    convey->chumax = convey->chunum = 0;
    convey->lastchunk = 0;
    convey->lastoperand = 0;
    convey->lastindex = 0;
    (void)inflateEnd(&convey->strm);
    bzero(&convey->strm, sizeof(z_stream));
out:
    ifBLIN_QX3("-");
    return;
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****  num-  (,                                     ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_virtualconvey(pngtr_conf *cf) {                                                            /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    int               ex    = EX_OK;

    ifBLIN_QX3("+ %08X %u", cf->flags, cf->num);
    PNGTR_X((cf->num >= PNGTR_CONVEYNUM),            ("Ill convey num %d", cf->num), ERANGE, EX_USAGE);
    pngtr_freefile(cf, 0);
    pngtr_freeconvey(cf);
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****     cf->num                                                ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_chudelete(pngtr_conf *cf) {                                                                /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey   *convey;
    u_int           top;
    int             ex = EX_OK;

    ifBLIN_QX3("+ %08X %u", cf->flags, cf->num);
    if  (!cf->convey || !(convey = &cf->convey[cf->num]) || !convey->chunks) goto out;
    top = convey->chunum;
    for (u_int chctr = 0, chto = 0; chctr <= convey->chunum; ++chctr) {
        pngtr_chunk *chunk;

    ;   ifBLIN_QX4("%u %u %u", chctr, chto, top);
    ;   chunk = &convey->chunks[chctr];
    ;   if  (  (chctr == convey->chunum)
            || !(  ((PNGTR_CHAINSL & convey->flags) && (PNGTR_SELECTD & chunk->flags))
                || ((PNGTR_CHATYPE & convey->flags) && ((1 << chunk->type) & convey->typemask))
                || !((PNGTR_CHAINSL | PNGTR_CHATYPE) & convey->flags)
            )   ) {
    ;       ifBLIN_QX4("%u %u", chctr, chto);
    ;       if  (chctr != chto) {
    ;           ifBLIN_QX4( "%"BLIN_X" %"BLIN_X" %u"
                          , chunk
                          , &convey->chunks[chto]
                          , sizeof(pngtr_chunk)
                          );
    ;           bcopy(chunk, &convey->chunks[chto], sizeof(pngtr_chunk));
    ;       }
    ;       ++chto;
    ;   } else {
    ;       pngtr_freechunk(cf, chctr);
    ;       --top;
    ;       ifBLIN_QX4("D %u %u %u", chctr, chto, top);
    }   }
    convey->chunum = top;
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****   png (mode == 0)  img (mode == 1)  num-                     ****
 ****      bos                                                              ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_read(pngtr_conf *cf, const char *o, int mode) {                                            /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey     *convey;
    int               fdin;
    int               ex    = EX_OK;
    pngtr_infile     *in;

    ifBLIN_QX3("+ %08X q%u %s=%s~", cf->flags, cf->num, !mode ? "png" : "img", o);
    PNGTR_X((cf->num >= PNGTR_CONVEYNUM),            ("Ill convey num %d", cf->num), ERANGE, EX_USAGE);
    if  (!mode) pngtr_freeconvey(cf); else pngtr_freefile(cf, 1);
    convey = &cf->convey[cf->num];
    in = !mode ? &convey->png : &convey->img;
    in->nm = strdup(o);
    PNGTR_W((0 > (fdin = open(o, O_RDONLY)))                           , ("fopen in %s", o), EX_IOERR);
    PNGTR_W(!(in->md = mife_init(((cf->flags >> 1) & BLIN_GEN1)))           , ("mife_init"), EX_IOERR);
    PNGTR_W((0 > mife_ctlfdsc(in->md, fdin))                  , ("mife_ctlfdsc(,%d)", fdin), EX_IOERR);
    PNGTR_W(!(in->cont = mife_window(in->md, 0, 0))                 , ("mife_window(,0,0)"), EX_IOERR);
    PNGTR_W((0 > (in->size = mife_ctlsize(in->md)))                      , ("mife_ctlsize"), EX_IOERR);
    convey->flags = PNGTR_CFTOCH & cf->flags;
out:
    if  (!!ex && in->md) {
        mife_fini(in->md);
        bzero(in, sizeof(pngtr_infile));
    }
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}
